/* Copyright (C) 1998 Free Software Foundation, Inc.

   This file is part of GNU Motti.

   GNU Motti is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Motti is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Motti; see the file COPYING.  If not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include "map.h"
#include "misc.h"
#include "snakes.h"

static int
readNumber(FILE *file2, int *status)
{
  int c;
  char buf[80];

  if(*status != 0)
    return 0;

  while(!isdigit(c=getc(file2)))
    {
      if(c == '\n')
	{
	  *status = 1;
	  return 0;
	}
    }
  ungetc(c, file2);

  return atoi(fgets(buf, 79, file2));
}

static void set_modes(Map *);

static int
get_char(FILE *f)
{
  int c;
  c = getc(f);
  if(c != '\n')
    ungetc(c, f);
  c = getc(f);
  if(c == EOF)
    return EOF;
  return c;
}

static void
free_map(Map *m)
{
  free(m->map);
  free(m->capital);
}

extern int
loadMap(char *mapDir, Map *map)
{
  FILE *file;
  int status = 0;
  int c;
  register int i;

  if(!(file = fopen(mapDir, "r")))
    return 0;

  map->width=readNumber(file, &status);
  map->height=readNumber(file, &status);
  map->warped=readNumber(file, &status);
  map->players=readNumber(file, &status);
  map->turn=readNumber(file, &status);
  map->n_att=readNumber(file, &status);

  if(status != 0)
    return 0;

  map->n_cross=0;
  map->bit_cross=0;

  map->size = map->width * map->height;
  map->map = (map_val *) mymalloc(map->size * sizeof(map_val));
  map->capital = (Coord *) mymalloc(map->players * sizeof(Coord));

  for(i = 0; i < map->size; i++)
    {
      c = get_char(file);
      if(c == EOF)
	{
	  free_map(map);
	  return 0;
	}
      if(islower(c))
	{
	  /* Number zero is reserved for sea */
	  map->map[i] = c-('a'-1);
	}
      else if(isupper(c))
	{
	  map->map[i] = (c-('A'-1)) | MASK_OCCUPIED;
	}
      else if(c == ' ')
	{
	  map->map[i] = 0;
	}
      else
	{
	  free_map(map);
	  return 0;
	}

      c = get_char(file);
      if(c == '.' || c == 'X')
	{
	  map->capital[map->map[i] & MASK_PLAYER].x = i % map->width;
	  map->capital[map->map[i] & MASK_PLAYER].y = i / map->width;
	  map->map[i] |= MASK_CAPITAL;
	}
      if(c == 'x' || c == 'X')
	{
	  map->cross[map->n_cross].x = i % map->width;
	  map->cross[map->n_cross++].y = i / map->width;
	  map->map[i] |= MASK_CROSS;
	  map->bit_cross |= 1;
	  map->bit_cross <<= 1;
	}
      if(c == EOF)
	{
	  free_map(map);
	  return 0;
	}
    }
  fclose(file);
  map->bit_cross >>= 1;
  set_modes(map);
  return 1;
}

extern void
get_map_real_coord(Coord *loc, Map map)
{
  if(!map.warped)
    {
      if(loc->x < 0 || loc->x > map.width || loc->y < 0 || loc->x >
	 map.height)
	{
	  loc->x = -1;
	  loc->y = -1;
	}
    }
  else
    {
      while(loc->x < 0)
	loc->x += map.width;
      while(loc->y < 0)
	loc->y += map.height;
      loc->x %= map.width;
      loc->y %= map.height;
    }
}

extern map_val
get_map_val(Map map, Coord loc)
{
  get_map_real_coord(&loc, map);

  if(loc.x == -1)
    return 0;
  return map.map[loc.x+map.width*loc.y];
}

extern Coord
add_coord(Coord a, Coord b)
{
  Coord ret_coord;
  ret_coord.x = a.x+b.x;
  ret_coord.y = a.y+b.y;
  return ret_coord;
}

extern int
cmp_coord(Map map, Coord a, Coord b)
{
  get_map_real_coord(&a, map);
  get_map_real_coord(&b, map);

  if(a.x == b.x && a.y == b.y)
    return 1;
  else
    return 0;
}

extern void
set_map(Map *map, Coord pos, map_val status)
{
  get_map_real_coord(&pos, *map);
  map->map[pos.x+pos.y*map->width] = status;
}

extern void
set_map_bit(Map *map, Coord pos, map_val status)
{
  get_map_real_coord(&pos, *map);
  map->map[pos.x+pos.y*map->width] |= status;
}

extern void
clear_map_bit(Map *map, Coord pos, map_val status)
{
  get_map_real_coord(&pos, *map);
  map->map[pos.x+pos.y*map->width] &= ~status;
}

static int
check_surrounding(Map *map, Coord loc, map_val me)
{
  unsigned register char i, j = 0;

  for(i = 0; i < 8; i++)
    {
      if(get_surrounding_val(map, loc, i) == me)
	{
	  /* Return one, if all modes possible, else two, for
	     guerilla.  */
	  if(i & 1)
	    return 1;
	  else
	    j = 1;
	}
    }
  if(j == 1)
    return 2;
  else
    return 0;
}

extern int
check_cross(Map *map, Coord loc)
{
  register int i;
  map_val myocc;

  if(map->n_cross == N_CROSS)
    {
      for(i = 0; i < N_CROSS; i++)
	{
	  if(cmp_coord(*map, map->cross[i], loc))
	    return 1;
	}
      return 0;
    }

  /* Check if a selected location is on own area.  */
  myocc = get_map_val(*map, loc) & MASK_PLAIN;
  if(myocc & MASK_OCCUPIED)
    return 0;
  if(myocc == map->turn && map->modes & (MODE_ATT | MODE_DEF))
    return 1;

  myocc = map->turn | MASK_OCCUPIED;
  if(map->n_cross <= 1)
    {
      return check_surrounding(map, loc, myocc);
    }
  else
    {
      for(i = 1; i < 8; i+=2)
	{
	  if(get_surrounding_val(map, loc, i) == myocc)
	    return 1;
	}
    }
  return 0;
}

extern int
toggle_cross(Map *map, Coord loc)
{
  unsigned register char i, j, k = N_CROSS+1, l;

  l = check_cross(map, loc);
  if(l)
    {
      for(i = 0, j = map->bit_cross; i < N_CROSS; i++, j >>= 1)
	{
	  if(j & 1)
	    {
	      if(cmp_coord(*map, map->cross[i], loc))
		{
		  map->bit_cross ^= 1<<i;
		  map->n_cross--;
		  set_map(map, loc, get_map_val(*map, loc) ^
			  MASK_CROSS);
		  set_modes(map);
		  return 1;
		}
	    }
	  else if(k == N_CROSS+1)
	    {
	      k = i;
	    }
	}
      if((l == 2 && map->n_cross == 1) || (l == 1 && !(map->modes &
						       (MODE_ATT|MODE_DEF))))
	{
	  return 0;
	}
      else
	{
	  map->bit_cross |= 1<<k;
	  map->n_cross++;
	  map->cross[k] = loc;
	  set_map(map, loc, get_map_val(*map, loc) | MASK_CROSS);
	  set_modes(map);
	  return 1;
	}
    }
  else
    return 0;
}

static void
set_modes(Map *map)
{
  int i;
  unsigned register char j, k;

  if(map->n_att == 1)
    {
      map->modes = MODE_ATT;
      map->def_mode = MODE_ATT;
    }
  else if(map->n_att == 0)
    {
      if(map->n_cross == 0)
	{
	  map->modes = MODE_ATT | MODE_DEF | MODE_GUE;
	  map->def_mode = MODE_ATT;
	}
      else if(map->n_cross > 3)
	{
	  map->modes = MODE_ATT;
	  map->def_mode = MODE_ATT;
	  return;
	}
      else if(map->n_cross > 1)
	{
	  for(j = map->bit_cross, k = 0; k < N_CROSS; k++, j >>= 1)
	    {
	      if(j & 1)
		{
		  if((get_map_val(*map, map->cross[k]) & MASK_PLAYER)
		     != map->turn)
		    {
		      map->modes = MODE_ATT;
		      map->def_mode = MODE_ATT;
		      return;
		    }
		}
	    }
	  map->modes = MODE_ATT | MODE_DEF;
	  map->def_mode = MODE_DEF;
	}
      else
	{
	  for(j = map->bit_cross, k = 0; k < N_CROSS; k++, j >>= 1)
	    {
	      if(j & 1)
		{
		  if((get_map_val(*map, map->cross[k]) & MASK_PLAYER)
		      == map->turn)
		    {
		      map->modes = MODE_ATT | MODE_DEF | MODE_GUE;
		      map->def_mode = MODE_ATT;
		    }
		  else
		    {
		      i = check_surrounding(map, map->cross[k],
					    map->turn |
					    MASK_OCCUPIED);
		      if(i == 1)
			{
			  map->modes = MODE_ATT | MODE_GUE;
			  map->def_mode = MODE_ATT;
			}
		      else
			{
			  map->modes = MODE_GUE;
			  map->def_mode = MODE_GUE;
			}
		    }
		  return;
		}
	    }
	}
    }
}

static int
check_cross_existence(Map *map, int cross_n)
{
  if((map->bit_cross >> cross_n) & 1)
    return 1;
  else
    return 0;
}

static void
occupy(Map *map, Action *act, int *cross_n)
{
  unsigned register int i, j, k;
  Region *areas;
  for(i = 0; i < act->count; i++)
    {
      map->bit_cross ^= 1<<*cross_n;
      map->n_cross--;
      set_map(map, map->cross[*cross_n], map->turn |
	      MASK_OCCUPIED);
      j = count_areas(map, map->cross[*(cross_n++)], &areas);
      if(j != 1)
	{
	  act->type |= MASK_EVENT_ENCIRCLEMENT;
	  for(k = 0; k < j; k++)
	    {
	      register int l;
	      unsigned register int p_bits = 0;
	      for(l = 0; l < map->players; l++)
		{
		  if(areas[k].players>>l ^ 1
		     && !find_capital(map, areas[k].begin, l))
		    p_bits |= 1<<l;
		}
	      if(p_bits != 0)
		encircled(map, act->loc[i], areas[k].begin, p_bits, act);
	    }
	}
    }
  free(areas);
}

extern void
action(Map *map, int action, Action *act)
{
  int i;

  if(action == 0)
    action = map->def_mode;

  if(action == MODE_ATT && map->modes & MODE_ATT)
    {
      act->type = EVENT_ATT;
      i = rand();
      i %= 6;
      if(check_cross_existence(map, i))
	{
	  int *cross_n;
	  cross_n = &i;
	  act->loc = (Coord *) mymalloc(sizeof(Coord));
	  *act->loc = map->cross[i];
	  act->count = act->base_count = 1;
	  occupy(map, act, cross_n);
	}
      else
	{
	  act->loc = NULL;
	  act->count = 0;
	}
    }
  else if(action == MODE_DEF && map->modes & MODE_DEF)
    {
      int cross_n_array[3];
      register int j;
      act->type = EVENT_DEF;
      act->loc = (Coord *) mymalloc(3*sizeof(Coord));
      for(i = 0, j = 0; i < N_CROSS; i++)
	{
	  if(check_cross_existence(map, i))
	    {
	      act->loc[j] = map->cross[i];
	      cross_n_array[j++] = i;
	    }
	}
      act->count = act->base_count = j;
      if(j)
	{
	  occupy(map, act, cross_n_array);
	}
      else
	{
	  free(act->loc);
	  act->loc = NULL;
	}
    }
  else if(action == MODE_GUE && map->modes & MODE_GUE)
    {
      act->type = EVENT_GUE;
      if(map->n_cross > 0)
	{
	  for(i = 0; i < N_CROSS; i++)
	    {
	      if(check_cross_existence(map, i))
		{
		  int *cross_n;
		  cross_n = &i;
		  act->loc = (Coord *) mymalloc(sizeof(Coord));
		  *act->loc = map->cross[i];
		  act->count = act->base_count = 1;
		  occupy(map, act, cross_n);
		  return;
		}
	    }
	  act->loc = NULL;
	  act->count = 0;
	}
    }
}

extern void
clean_map(Map *map, map_val m_val)
{
  register int i;
  
  m_val = ~m_val;
  for(i = 0; i < map->size; i++)
    map->map[i] &= m_val;
}
