/* Copyright (C) 1998 Free Software Foundation, Inc.

   This file is part of GNU Motti.

   GNU Motti is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Motti is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Motti; see the file COPYING.  If not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "xwin.h"
#include "map.h"

extern Window main_win;
extern Window but_win[NBUTS];
extern Window map_win;
extern Cursor active_cursor, disabled_cursor;

extern Display *display;
extern int screen_num;

extern void
eventLoop(XMapStruct *xmap)
{
  XEvent report;
  int x, y;
  Action map_events;

  XSelectInput(display, main_win, 0);
  XSelectInput(display, but_win[ATT_N], ExposureMask | ButtonPressMask);
  XSelectInput(display, but_win[DEF_N], ExposureMask | ButtonPressMask);
  XSelectInput(display, but_win[GUE_N], ExposureMask | ButtonPressMask);
  XSelectInput(display, map_win, ExposureMask | ButtonPressMask |
	       PointerMotionHintMask);

  XMapWindow(display, main_win);
  XMapWindow(display, but_win[ATT_N]);
  XMapWindow(display, but_win[DEF_N]);
  XMapWindow(display, but_win[GUE_N]);
  XMapWindow(display, map_win);

  while(1)
    {
      XNextEvent(display, &report);

      switch(report.type)
	{
	case Expose:
	  if(report.xexpose.window == map_win)
	    drawMap(report.xexpose.x, report.xexpose.y,
		    report.xexpose.width, report.xexpose.height, *xmap);

	  if(report.xexpose.count == 0)
	    {
	      if(report.xexpose.window == but_win[ATT_N])
		{
		  drawButton(ATT_BUT_MASK);
		}
	      else if(report.xexpose.window == but_win[DEF_N])
		{
		  drawButton(DEF_BUT_MASK);
		}
	      else if(report.xexpose.window == but_win[GUE_N])
		{
		  drawButton(GUE_BUT_MASK);
		}
	    }
	  break;
	case MotionNotify:
	  while(XCheckMaskEvent(display, ButtonMotionMask, &report));
	  if(!XQueryPointer(display, report.xmotion.window,
			    (Window *) NULL, (Window *) NULL, (int *)
			    NULL, (int *) NULL, &x, &y, (int *) NULL))
	    /* Pointer is on an other screen.  */
	    break;

	  if(check_cross(&xmap->map, win_pos2map_coord(xmap, x, y)))
	    XDefineCursor(display, map_win, active_cursor);
	  else
	    XDefineCursor(display, map_win, disabled_cursor);
	  break;
	case ButtonPress:
	  if(!report.xbutton.same_screen) break;
	  if(report.xbutton.window == but_win[ATT_N])
	    {
	      action(&xmap->map, MODE_ATT, &map_events);
	      draw_effects(*xmap, map_events);
	    }
	  else if(report.xbutton.window == but_win[DEF_N])
	    {
	      action(&xmap->map, MODE_DEF, &map_events);
	      draw_effects(*xmap, map_events);
	    }
	  else if(report.xbutton.window == but_win[GUE_N])
	    {
	      action(&xmap->map, MODE_GUE, &map_events);
	      draw_effects(*xmap, map_events);
	    }
	  else if(report.xbutton.window == map_win)
	    {
	      Coord coord;
	      coord = win_pos2map_coord(xmap, report.xbutton.x,
					report.xbutton.y);

	      if(toggle_cross(&xmap->map, coord))
		drawMap(report.xbutton.x, report.xbutton.y, 0, 0,
			*xmap);
	    }
	}
    }
}
