/* Copyright (C) 1998 Free Software Foundation, Inc.

   This file is part of GNU Motti.

   GNU Motti is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Motti is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Motti; see the file COPYING.  If not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* This file has general X init routines. Almost all of it is taken
 * from O'Reilly's Xlib Programming manual.
 */


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/xpm.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <getopt.h>

#include "xinit.h"
#include "misc.h"

Display *display;
char DisplayName[256] = {0};
int screen_num;
Colormap colormap;
int depth;

static Visual *visual;
static XrmDatabase parseDB;
#ifdef OLD_ARG_PROCESS
static const int opTableEntries = 4;
static XrmOptionDescRec opTable[4] = {
  {"-display", ".display", XrmoptionSepArg, (XPointer) NULL},
  {"-players", ".players", XrmoptionSepArg, (XPointer) NULL},
  {"-map",     ".map",     XrmoptionSepArg, (XPointer) NULL},
  {"-mono",    ".mono",    XrmoptionNoArg,  (XPointer) "1"}};
#endif

/* TODO: add more arguments. This needn't be such a mess. */
static void
process_x_args(int *argc, char *argv[])
{
  while(1)
    {
      int c;
      static struct option long_options[] =
      {
	{"display", 1, 0, 'd'},
	{"help", 0, 0, 'h'},
	{"version", 0, 0, 'v'},
	{0, 0, 0, 0}
      };
      
      c = getopt_long(*argc, argv, "d:", long_options, (int *) NULL);
      switch(c)
	{
	case EOF:
	  return;
	case 'd':
	  XrmPutStringResource(&parseDB, "xmotti.display", optarg);
	  break;
	}
    }
}
      

/* This function might seem cryptical, but it just determines which
 * display to use and parses the command line to determine it. */
extern void
openDisplay(int *argc, char *argv[])
{
  XrmValue value;
  char *str_type[20];
  XrmDatabase displayDB = {0};

  //  DisplayName[0] = '\0';
#ifdef OLD_ARG_PROCESS
  XrmParseCommand(&parseDB, opTable, opTableEntries, argv[0], argc,
		  argv);
#endif
  process_x_args(argc, argv);
  XrmMergeDatabases(parseDB, &displayDB);

  /*  if(*argc != 1) usage();*/

  /* These are used to search the display value in local file
   * databases. I have no idea if I search from the correct locations.
   */
  XrmCombineFileDatabase("~/.Xdefaults", &displayDB, False);
  XrmCombineFileDatabase("/usr/lib/X11/app-defaults/XMotti",
			 &displayDB, False);

  if(XrmGetResource(displayDB, "xmotti.display", "XMotti.Display",
		    str_type, &value))
    {
      strncpy(DisplayName, value.addr, (int) value.size);
    }

  if(!(display = XOpenDisplay(DisplayName)))
    {
      fprintf(stderr, "%s: Can't open display: %s\n", argv[0],
	      XDisplayName(DisplayName));
      exit(1);
    }

  /* These are defined now, altough they'll be needed only later... */
  screen_num = DefaultScreen(display);
  visual = DefaultVisual(display, screen_num);
  colormap = DefaultColormap(display, screen_num);
  depth = DefaultDepth(display, screen_num);

  XrmDestroyDatabase(displayDB);
}

XrmDatabase DB;

void
createFinalDB()
{
  XrmDatabase appDB, homeDB, serverDB;

  //  char *str_type[20];
  char filenamebuf[1024];
  char *filename = &filenamebuf[0];
  char *environment;
  char classname[]="XMotti";

  strcpy(filename, "/usr/lib/X11/app-defaults/");
  strcat(filename, classname);

  appDB = XrmGetFileDatabase(filename);
  XrmMergeDatabases(appDB, &DB);

  /* I don't like the way this is made, but I just copied it from
   * O'Reilly. I'm innocent. */
  if(XResourceManagerString(display) != NULL)
    {
      serverDB = XrmGetStringDatabase(XResourceManagerString(display));
    }
  else
    {
      environment = getenv("HOME");
      strcpy(filename, environment);
      strcat(filename, "/.Xdefaults");

      serverDB = XrmGetFileDatabase(filename);
    }
  XrmMergeDatabases(serverDB, &DB);
  
  if((environment = getenv("XENVIRONMENT")) == NULL)
    {
      int len;
      environment = getenv("HOME");
      strcat(environment, "/.Xdefaults-");
      len = strlen(environment);
      gethostname(environment+len, 1024-len);
    }
  homeDB = XrmGetFileDatabase(environment);

  XrmMergeDatabases(homeDB, &DB);
  XrmMergeDatabases(parseDB, &DB);
}

XColor *player_col;
XColor *occupied_col;
XColor capital_col, cross_col;
XColor bg_col, fg_col;
XColor map_bg_col;
int players;
Bool color;
/* Note to myself: there's no BW support yet. Do it. */

static void
cantAllocColor()
{
  fprintf(stderr, "Can't alloc color!\n");
  color = False;
}

static Status
allocXrmColor(char *instance, char *class, XColor *color)
{
  char *str_type[20];
  XrmValue value;
  XColor nullcol;

  if(XrmGetResource(DB, instance, class, str_type, &value) == False)
    {
      fprintf(stderr, "Can't find color for %s\n", instance);
      cantAllocColor();
      return 0;
    }
  if(!XAllocNamedColor(display, colormap, value.addr, color, &nullcol))
    {
      fprintf(stderr, "Can't allocate color %s\n", value.addr);
      cantAllocColor();
      return 0;
    }
  return 1;
}

extern void
allocColors()
{
  long plane_mask[1];
  long *pixels;
  int npixels;
  char *str_type[20];
  XrmValue value;
  int i;
  char instancename[256];
  XVisualInfo visual_info;

  /* Search for the best visual, use color if possible */
  for(i = 5; !XMatchVisualInfo(display, screen_num, depth, i--,
			       &visual_info););
  if(i < StaticColor)
    {
      cantAllocColor();
      return;
    }

  color = True;

  if(XrmGetResource(DB, "xMotti.players", "XMotti.Players", str_type,
		    &value))
    {
      players = atoi(value.addr);
    }
  else
    {
      players = 2;
    }

  if(players > MAXPLAYERS)
    {
      fprintf(stderr, "Too many players!\n");
      exit(1);
    }
  else if(players < 2)
    {
      fprintf(stderr, "Too few players!\n");
      exit(1);
    }

  player_col = (XColor *) mymalloc(sizeof(XColor)*players);
  occupied_col = (XColor *) mymalloc(sizeof(XColor)*players);

  /* Determine if there is enough color cells available for all
   * colors. The game needs 2 colors per player and one for capital,
   * one for cross,  one for sea, one for background and one for
   * foreground. The amount of allocated colorcells might be lower. */
  pixels = (long *) mymalloc(sizeof(long)*(npixels = players*2+4));

  if(!XAllocColorCells(display, colormap, False, plane_mask, 0,
		       pixels, npixels))
    {
      cantAllocColor();
      return;
    }
  XFreeColors(display, colormap, pixels, npixels, 0);

  if(!allocXrmColor("xMotti.background", "XMotti.Background",
		    &bg_col))
    return;
  if(!allocXrmColor("xMotti.foreground", "XMotti.Foreground",
		    &fg_col))
    return;

  for(i = 1; i <= players; i++)
    {
      sprintf(instancename, "xMotti.player%i.color", i);

      if(!allocXrmColor(instancename, "XMotti.Player.Color",
			&player_col[i-1]))
	return;

      sprintf(instancename, "xMotti.player%iOcc.color", i);

      if(!allocXrmColor(instancename, "XMotti.PlayerOcc.Color",
			&occupied_col[i-1]))
	return;
    }

  if(!allocXrmColor("xMotti.color.map.background",
		    "XMotti.Map.Background", &map_bg_col))
    return;

  if(!allocXrmColor("xMotti.color.capital", "XMotti.Color.Capital",
		&capital_col))
    return;

  allocXrmColor("xMotti.color.cross", "XMotti.Color.Cross",
		&cross_col);
}

static void
makeMaskStipples(Pixmap capital, Pixmap cross, int width, int height)
{
  GC tempgc;
  XGCValues gcvalues = {0};

  gcvalues.line_width = 1;
  tempgc = XCreateGC(display, capital, GCLineWidth, &gcvalues);

  XSetForeground(display, tempgc, 1);
  XFillRectangle(display, capital, tempgc, 0, 0, width+1, height+1);
  XSetForeground(display, tempgc, 0);
  XFillRectangle(display, capital, tempgc, width/3, height/3,
		 2*width/3, 2*height/3);

  XFillRectangle(display, cross, tempgc, 0, 0, width+1, height+1);
  XSetForeground(display, tempgc, 1);
  XDrawLine(display, cross, tempgc, width/6, height/6, (5*width)/6,
	    (5*height)/6);
  XDrawLine(display, cross, tempgc, (5*width)/6, height/6, width/6,
	    (5*height)/6);

  XFreeGC(display, tempgc);
}

Window main_win;
Window but_win[NBUTS];
Window map_win;

GC maingc, mapgc, capitalgc, crossgc/*, effectgc*/;

/* GC and pixmap for disabled action buttons. */
GC dgc;
Pixmap dpixmap;

static void
alloc_gcs(XGCValues *gcval)
{
  Pixmap capital_stipple, cross_stipple;

  if(!(capital_stipple=XCreatePixmap(display, main_win, MAPSQUARESIZE,
				     MAPSQUARESIZE, 1)) ||
     !(cross_stipple=XCreatePixmap(display, main_win, MAPSQUARESIZE,
				   MAPSQUARESIZE, 1)))
    {
      fprintf(stderr, "Cannot allocate pixmap!\n");
      exit(1);
    }

  makeMaskStipples(capital_stipple, cross_stipple, MAPSQUARESIZE,
		   MAPSQUARESIZE);

  /* These three are set to draw capitals. */
  gcval->foreground = capital_col.pixel;
  gcval->fill_style = FillStippled;
  gcval->stipple = capital_stipple;
  capitalgc = XCreateGC(display, main_win, GCForeground | GCFillStyle
			| GCStipple, gcval);
  gcval->foreground = cross_col.pixel;
  gcval->stipple = cross_stipple;
  crossgc = XCreateGC(display, main_win, GCForeground | GCFillStyle |
		      GCStipple, gcval);


  /* No fields need to be set, since foreground color is changed every
   * time */
  /* Note to myself: change this to make BW window. */
  mapgc = XCreateGC(display, main_win, 0, gcval);

  gcval->foreground = fg_col.pixel;
  gcval->background = bg_col.pixel;
  gcval->font = XLoadFont(display, "9x15");
  maingc = XCreateGC(display, main_win, GCForeground | GCBackground |
		     GCFont, gcval);

  /* This is the gc used to show a disabled action button */
  gcval->fill_style = FillStippled;
  gcval->stipple = dpixmap;
  dgc = XCreateGC(display, main_win, GCForeground | GCBackground |
		  GCFillStyle | GCStipple, gcval);

  /* This gc is used to show that some place is being occupied */
  /*effectgc = XCreateGC(display, main_win, GCFillStyle | GCStipple,
    &gcval);*/
  /* Note to myself: this could be very complex. */
}

#include "bitmaps/disabled"

/* This one is initialised in openWindows and used also in printStatus */
XFontStruct *fontstruct;

int but_width[3];
int but_height;

/* Note to myself: this is a mess! It should be divided into several
 * functions to get it more logical. */
/* Note to myself: change this to implement monochrome display. */
extern void
openWindows(int argc, char **argv, XMapStruct *xmap)
{
  XrmValue value;
  XSetWindowAttributes win_attribs;
  int x = 0, y = 0, width, height;
  int start, end;
  char *str_type[20];
  XGCValues gcvalues = {0};

  XWMHints *wm_hints;
  XClassHint *class_hints;
  XTextProperty windowName, iconName;
  char *window_name = "XMotti";
  char *icon_name = "XMotti";
  XSizeHints *size_hints;

  /* Allocate needed structures for properties */
  if(!(size_hints=XAllocSizeHints()) || !(wm_hints = XAllocWMHints())
     || !(class_hints=XAllocClassHint()) ||
     !XStringListToTextProperty(&window_name, 1, &windowName) ||
     !XStringListToTextProperty(&icon_name, 1, &iconName))
    {
      fprintf(stderr, "Cannot allocate memory!\n");
      exit(0);
    }

  /* Note to myself: read the O'Reilly a few more times to understand
   * these better. Buy more books. */
  XrmGetResource(DB, "XMotti.Geometry", "xMotti.geometry", str_type,
		 &value);
  XParseGeometry(value.addr, &x, &y, &width, &height);

  win_attribs.background_pixel = bg_col.pixel;
  main_win = XCreateWindow(display, RootWindow(display, screen_num),
			   x, y, width, height, 0, CopyFromParent,
			   InputOutput, CopyFromParent, CWBackPixel,
			   &win_attribs);

  /* This creates the pixmap used for disabled action button */
  if(!(dpixmap=XCreateBitmapFromData(display, main_win, disabled_bits,
				     disabled_width,
				     disabled_height)))
    {
      fprintf(stderr, "Cannot create pixmap!\n");
      exit(1);
    }

  alloc_gcs(&gcvalues);

  fontstruct = XQueryFont(display, gcvalues.font);
  start = X_MARGIN;
  end = XTextWidth(fontstruct, ATT_MSG, strlen(ATT_MSG))+X_MARGIN;
  but_width[ATT_N] = end-start;
  height = fontstruct->ascent+fontstruct->descent;
  but_height = height+6;

  but_win[ATT_N] = XCreateSimpleWindow(display, main_win, start,
				    Y_MARGIN, end, but_height, 2,
				    fg_col.pixel, bg_col.pixel);

  start = (end += X_MARGIN);
  end += XTextWidth(fontstruct, DEF_MSG, strlen(DEF_MSG))+X_MARGIN;
  but_width[DEF_N] = end-start;
  but_win[DEF_N] = XCreateSimpleWindow(display, main_win, start,
				    Y_MARGIN, end, but_height, 2,
				    fg_col.pixel, bg_col.pixel);
  start = (end += X_MARGIN);
  end += XTextWidth(fontstruct, GUE_MSG, strlen(GUE_MSG))+X_MARGIN;
  but_width[GUE_N] = end-start;
  but_win[GUE_N] = XCreateSimpleWindow(display, main_win, start,
				    Y_MARGIN, end, but_height, 2,
				    fg_col.pixel, bg_col.pixel);

  /* The map should be actually placed in the lower right corner to
   * use SouthWestGravity, but it doesn't use it now. Note to myself:
   * do it. */
  x = X_MARGIN;
  /* There must be room for action buttons and status message above
   * the map */
  y = Y_MARGIN*2 + height + but_height;
  /* Note to myself: the map size is presently fixed. Implement
   * resizeable map. */
  map_win = XCreateSimpleWindow(display, main_win, x, y,
				xmap->map.width * MAPSQUARESIZE,
				xmap->map.height * MAPSQUARESIZE, 2,
				fg_col.pixel, map_bg_col.pixel);

  /* Note to myself: do this correctly */
  size_hints->flags = PPosition | PSize | PMinSize;
  size_hints->min_width = end + X_MARGIN;
  size_hints->min_height = y + MINMAPHEIGHT*MAPSQUARESIZE + Y_MARGIN;

  wm_hints->initial_state = NormalState;
  wm_hints->input = True;
  
  /* Note to myself: do this correctly */
  class_hints->res_name = "xmotti";
  class_hints->res_class = "XMotti";

  XSetWMProperties(display, main_win, &windowName, &iconName, argv,
		   argc, size_hints, wm_hints, class_hints);

  /* Free all allocated structures */
  XFree(size_hints);
  XFree(wm_hints);
  XFree(class_hints);
  XFree(windowName.value);
  XFree(iconName.value);
}

#include "bitmaps/map_pointer_active"
#include "bitmaps/map_pointer_disabled"
#include "bitmaps/map_pointer_mask"
Cursor active_cursor, disabled_cursor;

extern void
create_cursors()
{
  Pixmap mask, active, disabled;
  XColor white, black;
  Colormap default_colormap;

  default_colormap = DefaultColormap(display, screen_num);
  white.pixel = WhitePixel(display, screen_num);
  black.pixel = BlackPixel(display, screen_num);
  XQueryColor(display, default_colormap, &white);
  XQueryColor(display, default_colormap, &black);

  mask = XCreateBitmapFromData(display, main_win,
			       map_pointer_mask_bits,
			       map_pointer_mask_width,
			       map_pointer_mask_height);
  active = XCreateBitmapFromData(display, main_win,
				 map_pointer_active_bits,
				 map_pointer_active_width,
				 map_pointer_active_height);
  disabled = XCreateBitmapFromData(display, main_win,
				   map_pointer_disabled_bits,
				   map_pointer_disabled_width,
				   map_pointer_disabled_height);

  active_cursor = XCreatePixmapCursor(display, active, mask, &black,
				      &white,
				      map_pointer_active_x_hot,
				      map_pointer_active_y_hot);
  disabled_cursor = XCreatePixmapCursor(display, disabled, mask,
					&black, &white,
					map_pointer_disabled_x_hot,
					map_pointer_disabled_y_hot);

  XFreePixmap(display, mask);
  XFreePixmap(display, active);
  XFreePixmap(display, disabled);
}
