/* Copyright (C) 1998 Free Software Foundation, Inc.

   This file is part of GNU Motti.

   GNU Motti is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Motti is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Motti; see the file COPYING.  If not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* This file has all the drawing routines */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "xwin.h"
#include "map.h"

extern Display *display;
extern XrmDatabase DB;
extern XColor bg_col, fg_col;
extern XColor *player_col, *occupied_col, map_bg_col, capital_col,
  cross_col;
extern int screen_num;

extern Window main_win;
extern Window but_win[NBUTS];
extern Window map_win;

extern GC maingc, mapgc, capitalgc, crossgc/*, effectgc*/;

/* This one is initialised in openWindows and used also in printStatus */
extern XFontStruct *fontstruct;

/* GC and pixmap for disabled action buttons. */
extern GC dgc;
extern Pixmap dpixmap;

extern int but_width[3];
extern int but_height;

extern void
drawButton(char buttonDrawMask)
{
  static int att_msg_len = 0, def_msg_len = 0, gue_msg_len = 0;
  static int offset = 0;

  if(offset == 0)
    offset = fontstruct->ascent + 3;

  if(buttonDrawMask & ATT_BUT_MASK)
    {
      if(att_msg_len == 0)
	att_msg_len = strlen(ATT_MSG);
      XDrawString(display, but_win[ATT_N], maingc, offset, 0, ATT_MSG,
		  att_msg_len);
    }
  if(buttonDrawMask & DEF_BUT_MASK)
    {
      if(def_msg_len == 0)
	def_msg_len = strlen(DEF_MSG);
      XDrawString(display, but_win[DEF_N], maingc, offset, 0, DEF_MSG,
		  def_msg_len);
    }
  if(buttonDrawMask & GUE_BUT_MASK)
    {
      if(gue_msg_len == 0)
	gue_msg_len = strlen(GUE_MSG);
      XDrawString(display, but_win[GUE_N], maingc, offset, 0, GUE_MSG,
		  gue_msg_len);
    }
}

static void
showDisabledButton(char buttonMask)
{
  if(buttonMask & ATT_BUT_MASK)
    XFillRectangle(display, but_win[ATT_N], dgc, 0, 0, but_width[ATT_N]+1,
		   but_height+1);

  if(buttonMask & DEF_BUT_MASK)
    XFillRectangle(display, but_win[DEF_N], dgc, 0, 0, but_width[DEF_N]+1,
		   but_height+1);

  if(buttonMask & GUE_BUT_MASK)
    XFillRectangle(display, but_win[GUE_N], dgc, 0, 0, but_width[GUE_N]+1,
		   but_height+1);
}

extern void
setDisabledButtons(char buttonMask)
{
  static char lastButtonMask = 0;
  char drawMask = 0;
  int mask = 0;
  unsigned register char i;

  for(i = 0; i <= 2; i++)
    {
      mask = 1L<<i;

      /* The button wasn't previously disabled and becomes now
       * disabled.  Draw disabled background and set window background
       * accordingly */
      if(!(lastButtonMask & mask) && buttonMask & mask)
	{
	  drawMask |= mask;
	  XSetWindowBackgroundPixmap(display, but_win[i], dpixmap);
	}

      /* The button becomes active */
      if(lastButtonMask & mask && !(buttonMask & mask))
	XSetWindowBackgroundPixmap(display, but_win[i], dpixmap);
    }
  
  if(drawMask != 0)
    showDisabledButton(drawMask);

  lastButtonMask = buttonMask;
}

static void
draw_map_function(XMapStruct xmap, Coord loc, Coord dim)
{
  Coord offset;
  short basex, lastx;
  int color;
  char lastStatus, status;

  basex = loc.x;

  if(dim.x == loc.x)
    dim.x++;
  for(offset.y = loc.y; offset.y <= dim.y; offset.y++)
    {
      lastx = basex;
      lastStatus = get_map_val(xmap.map, add_coord(loc, xmap.offset));

      for(offset.x = loc.x; offset.x <= dim.x; offset.x++)
	{
	  status = get_map_val(xmap.map, add_coord(offset, xmap.offset));
	  if(lastStatus != status || offset.x == dim.x)
	    {
	      color = (lastStatus & MASK_OCCUPIED ?
		       occupied_col[(lastStatus &
				     MASK_PLAYER)-1].pixel :
		       lastStatus ? player_col[(lastStatus &
						MASK_PLAYER)-1].pixel :
		       map_bg_col.pixel);
	      XSetForeground(display, mapgc, color);

	      XFillRectangle(display, map_win, mapgc, lastx *
			     MAPSQUARESIZE, offset.y * MAPSQUARESIZE,
			     (offset.x-lastx) * MAPSQUARESIZE,
			     MAPSQUARESIZE);

	      if(lastStatus & MASK_CAPITAL)
		{
		  XFillRectangle(display, map_win, capitalgc, lastx *
				 MAPSQUARESIZE, offset.y *
				 MAPSQUARESIZE, (offset.x-lastx) *
				 MAPSQUARESIZE, MAPSQUARESIZE);
		}
	      if(lastStatus & MASK_CROSS)
		{
		  XFillRectangle(display, map_win, crossgc, lastx *
				 MAPSQUARESIZE, offset.y *
				 MAPSQUARESIZE, (offset.x-lastx) *
				 MAPSQUARESIZE, MAPSQUARESIZE);
		}
	      lastx = offset.x;
	      lastStatus = status;
	    }
	}
    }
}

extern void
drawMap(int x, int y, int width, int height, XMapStruct xmap)
{
  Coord loc, dim;

  loc.x = x / MAPSQUARESIZE;
  loc.y = y / MAPSQUARESIZE;
  dim.x = (width % MAPSQUARESIZE ? width / MAPSQUARESIZE + 1 :
	      width / MAPSQUARESIZE)+loc.x; 
  dim.y = (height % MAPSQUARESIZE ? height / MAPSQUARESIZE + 1 :
	       height / MAPSQUARESIZE)+loc.y;

  draw_map_function(xmap, loc, dim);
}

extern void
draw_effects(XMapStruct xmap, Action map_events)
{
  register int i;
  for(i = 0; i < map_events.count; i++)
    {
      Coord null = {0};
      draw_map_function(xmap, map_events.loc[i], null);
    }
  free(map_events.loc);
}

extern Coord
win_pos2map_coord(XMapStruct *xmap, int x, int y)
{
  Coord ret_coord;
  ret_coord.x = x / MAPSQUARESIZE + xmap->offset.x;
  ret_coord.y = y / MAPSQUARESIZE + xmap->offset.y;
  return ret_coord;
}
